#!/usr/bin/env python

__author__ = "Philipp Blandfort"


#===========================================================================================================================
# IMPORTS
#===========================================================================================================================

import re
from scipy import *
from numpy import *
from matplotlib.pyplot import *
from scipy.ndimage.filters import gaussian_filter as gf
from scipy.ndimage.filters import laplace
from scipy.ndimage.filters import maximum_filter
from scipy.ndimage.filters import minimum_filter
from scipy.ndimage.filters import median_filter



#===========================================================================================================================
# CONSTANTS
#===========================================================================================================================

sampling_rate = 10000


#===========================================================================================================================
# FUNCTIONS
#===========================================================================================================================

def readData(filename = '../data/130207 E3 ChalXHzRec1Hz.txt'):
    '''Read all experimental data from a .txt file.'''
    f = open(filename, 'r')
    data = []
    
    # get the column number (-> number of different frequencies) from the document
    firstline = f.readline()
    for i in range(len(firstline.split())):
        data.append([])
    
    # read the data
    for line in f:
        words = line.split()
        
        if len(words)!=len(data):
            print 'warning: different column number for line "',line,'" (line skipped)'
            continue
        for i in range(len(words)):
            if words[i]=='0':
                data[i].append(0.)
                continue
            
            m = re.search('^(-?[0-9]+),([0-9]*)E(-?[0-9]+)$', words[i])
            
            if m:
                data[i].append(float(m.group(1) + '.' + m.group(2)) * 10**(int(m.group(3)) + 12)) # +12 to get data in pA
            else:
                m = re.search('^(-?[0-9]+)E(-?[0-9]+)$', words[i])
                if m:
                    data[i].append(float(m.group(1)) * 10**(int(m.group(2)) + 12)) # +12 to get data in pA
        
    #TODO: check if in each line all numbers have been extracted correctly
    f.close()
    for i in range(len(data)):
        data[i] = array(data[i])
    data = array(data)
    return data


def release(data, thr=30, t=4):
    '''Return an array r where r[t] is 1 if data[t] belongs to a release and 0 else.

    It is assumed that releases can be characterized as places where the current is
    above some threshold thr for some duration t.'''
    elevated = (gf(data,1) > thr) * 1.
    # Note: A light Gaussian filter is applied to make the method more robust.
    
    # Note: The following could be done in a slightly more elegant way with morphology. (Using opening I think.)
    rightE = copy(elevated)
    for i in range(1,t+1):
        rightE *= roll(elevated, -i)
    leftE = copy(elevated)
    for i in range(1,t+1):
        leftE *= roll(elevated, i)
    return (((rightE * 1.) + (leftE * 1.)) >= 1) * 1.


def artifacts(data, thr_u = 30, thr_l = -20, N = 10, thr_laplace = 80, **release_args):
    '''Detect artifacts.

    Keyword arguments:
    thr_u --- the upper threshold for artifact detection
    thr_l --- the lower threshold for artifact detection
    thr_laplace --- the threshold for the absolute value of the Laplace filter
    N --- '''
    candidates = ((data > thr_u) + (data < thr_l)) * (1 - release(data, **release_args))
    # those are places where the current I is especially low or high and there is no release
    
    # now combine that with Laplace
    lapl = abs(laplace(data) * candidates) # abs() since artifacts might be spikes pointing downward or upward
    return (lapl>thr_laplace) * (lapl>=maximum_filter(lapl,N))


def baselineCurrent(data):
    '''Compute the baseline current of the given data.

    This function should be applied to the full data of one run since it is assumed that there is
    no stimulus during the first 10k samples.'''
    return mean(data[:9900])


def baselineCorrection(data):
    '''Subtract the baseline current of the given data and return the normalized data.

    This function should be applied to the full data of one run since it is assumed that there is
    no stimulus during the first 10k samples.'''
    return data - baselineCurrent(data)


def stimulus(data, end_without_stimulus=25000, display_warnings=True, frequency=1., r=50):
    '''Return an array s where s[t] is 1 if there is a stimulus at time t and 0 else.

    Keyword arguments:
    frequency --- the stimulus frequency (during first minute) in Hz'''
    artis = artifacts(baselineCorrection(data), thr_laplace=10)
    stimulus = zeros(len(data))
    
    cur_pos = sampling_rate
    stepsize = int(sampling_rate*1./frequency)
    stimulus_counter = 0
    
    while cur_pos < len(data) - end_without_stimulus:
        # find artifact in [cur_pos-r,cur_pos+r]
        cur_arti = -1
        for i in range(r):
            if artis[cur_pos-i]:
                cur_arti = cur_pos - i
                break
            if cur_pos+i < len(data) and artis[cur_pos+i]:
                cur_arti = cur_pos + i
                break
        # if no artifact was found in [cur_pos-r,cur_pos+r]
        if cur_arti == -1:
            if display_warnings:
                print 'warning: failure at',cur_pos
            # use the estimation as stimulus time
            cur_arti = cur_pos
        stimulus[cur_arti] = 1.
            
        stimulus_counter += 1
        cur_pos = cur_arti + stepsize

        # the frequency changes back to 1Hz after 60s of stimulus!
        #TODO this doesn't work for 200Hz! (maybe use release detection information to get transition right)
        if stimulus_counter >= 60*frequency: #cur_pos > 61 * sampling_rate:
            stepsize = sampling_rate
        # Remark:
        # Due to drifting, the transition might be a little later or earlier than 61 * sampling_rate.
        # However, it is quite important to get the transition right since an offset might mess up later detection.
    return stimulus


def removeNoise(data, b=6, smoothing=1):
    '''Remove noise from experimental data.

    The noise is removed by applying a median filter followed by some Gaussian smoothing.

    Keyword arguments:
    data --- Output data of an experiment as array.
    b --- size of median filter
    smoothing --- amount of Gaussian smoothing'''
    if smoothing>0:
        return gf(median_filter(data, b), smoothing)
    else:
        return median_filter(data, b)


#===========================================================================================================================
# USAGE
#===========================================================================================================================

#data = readData()
#frequencies = [1, 2, 5, 10, 50, 100, 200]

#for i in range(len(data)):
#    np.savez('data/'+str(frequencies[i])+'Hz.npz', stimulus=stimulus(data[i], frequency=frequencies[i]),\
#        output=baselineCorrection(removeNoise(data[i])), raw=data[i])
